/*
** Module   : MenuContainerWidget.java
** Abstract : The base class for MenuWidget and SubMenuWidget.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 VIG 20150115 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150304 Added firstChild method, id's collision fix and other minor changes
*/
package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;

/**
 * This class represents sub-menu widget.
 */
abstract class MenuContainerWidget<T extends WidgetConfig>
extends GenericWidget<T> 
{
   /** Maps widget names to widget instances (from the menu definition). */
   private Map<String, GenericWidget<?>> n2w = null;

   /** List of widgets in ID order. */
   private List<GenericWidget<?>> widgets = null;
   
   /** menu definition */
   private MenuDescription menuDef = null;
   
   /** Shared menu flag. */
   private boolean shared = false;

   /** batch modifications flag */
   private boolean batch = true;
   
   /** PARENT attribute */
   private MenuContainerWidget<T> parent = null;

   /**
    * Default constructor.  
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    menuConfig
    *           The menu config instance.
    */
   public MenuContainerWidget(boolean dynamic, T menuConfig)
   {
      super(dynamic, menuConfig);
   }

   /**
    * Check if menu is shared.
    * 
    * @return  <code>true</code> if menu is shared.
    */
   public boolean isShared()
   {
      return shared;
   }
   
   /**
    * Set shared flag for menu.
    * 
    * @param   shared
    *          <code>true</code> if menu is shared.
    */
   void setShared(boolean shared)
   {
      this.shared = shared;
   }
   
   /**
    * Sets batching mode for frame modification.
    *
    * @param  on
    *         <code>true</code> sets batching mode
    */
   public void batch(boolean on)
   {
      batch(on, false);
   }

   /**
    * Queries batching mode for frame modification.
    *
    * @return 
    *         <code>true</code> if batching mode is in effect
    */
   public boolean isBatch()
   {
      return batch;
   }

   /**
    * Sets batching mode for frame modification and optionally pushes the
    * screen definition to the client. 
    *
    * @param  on
    *         <code>true</code> sets batching mode
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push. This is only honored when <code>on</code> is
    *         <code>false</code>.
    */
   public void batch(boolean on, boolean force)
   {
      batch = on;
      
      if (!batch)
         pushMenuDescription(force);
   }
   
   /**
    * Gets the parent widget - MENU or SUB-MENU
    *
    * @return   the parent widget for this container.
    */
   @Override
   @SuppressWarnings("unchecked")
   public CommonWidget getParent()
   {
      if (parent == null && this instanceof SubMenuWidget)
      {
         SubMenuConfig cfg = (SubMenuConfig) config;
         parent = (MenuContainerWidget<T>) LogicalTerminal.getWidgetForId(cfg.parentId);
      }
      
      return parent;
   }

   /**
    * Get all widgets stored in the container.
    *
    * @return  array of stored widgets.
    */
   public Collection<GenericWidget<?>> widgets()
   {
      return Collections.unmodifiableCollection(widgets);
   }
   
   /**
    * Core initialization processing. 
    *
    * @param    defClass
    *           The class of the definition implementation class.
    * @param    defInstance
    *           The instance of the definition implementation class.
    * @param    dynamic
    *           <code>true</code> If this menu is dynamic, <code>false</code> otherwise.
    */
   protected void coreInitialize(Class<?>         defClass,
                               Object             defInstance, 
                               boolean            dynamic)
   {
      WidgetListBase<?> wl = (WidgetListBase<?>) defInstance;
 
      // cache the widget map for this menu
      n2w = wl.getWidgets();
      
      // init the widget array
      widgets = new ArrayList<GenericWidget<?>>(n2w.size());
      
      // dynamic menu initialization should stop here
      if (dynamic)
      {
         return;
      }

      // add all the widgets in the order of their definition
      Iterator<String> iter = n2w.keySet().iterator();
      
      while (iter.hasNext())
      {
         
         // get the widget
         GenericWidget<?> widget = n2w.get(iter.next()); 
         widgets.add(widget);
         
         if (widget.getId() == -1) // widget instanceof SubMenuWidget
         {
            int wid = WidgetId.nextID().asInt();

            // initialize key state in the widget
            widget.setId(wid);

            // save new widget in session registry map
            LogicalTerminal.registerWidget(widget, wid);
         }
      }
   }
   
   /**
    * Conditionally pushes the screen definition instance down to the client. 
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of 
    * <code>LogicalTerminal</code> is called to send it down to the client.
    * 
    * Overrides {@link GenericWidget#pushScreenDefinition()} behavior.
    */
   protected void pushScreenDefinition()
   {
      pushMenuDescription(true);
   }
   
   /**
    * Conditionally pushes the screen definition instance down to the client. 
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of 
    * <code>LogicalTerminal</code> is called to send it down to the client.
    * 
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push.
    */
   void pushMenuDescription(boolean force)
   {
      // batching is in progress, do not push individual changes
      if (batch)
         return;
      
      CommonWidget parent = getParent();

      // time to push the changes down
      if (menuDef == null)
      {
         // create a new menuDef
         int menuId = config.id.asInt();
         menuDef = new MenuDescription(menuId);
         menuDef.addConfig(config, menuId);

         // add all the widgets in the order of their definition
         
         for (GenericWidget<?> gw : widgets)
         {
            menuDef.addConfig(gw.config());
            if (gw instanceof SubMenuWidget)
            {
               MenuContainerWidget<?> sm = (MenuContainerWidget<?>) gw;
               sm.pushMenuDescription(force);
               menuDef.addChildDescription(sm.menuDef, sm.config.id.asInt());
            }
         }
         
         if (isShared() && isVisible().booleanValue())
         {
            menuDef.setSuppressRedraw(true);
         }
      }
      
      // push down the screen definition if this menu is top level
      if (parent == null || !(parent instanceof MenuContainerWidget<?>))
      {
         if (force)
         {
            // synchronous push
            MenuDescription[] sd = new MenuDescription[] {menuDef};
            LogicalTerminal.pushMenuDescription(sd);
         }
         else
         {
            // deferred push
            LogicalTerminal.deferPush(menuDef);
         }
      }
      else
      {
         ((MenuContainerWidget<?>)parent).pushMenuDescription(force);
      }
   }
   
   /**
    * Finish the config processing for this frame.
    * This includes registering the active configuration for each new widget.
    */
   void finishConfigProcessing()
   {
      ConfigManager mgr = ConfigManager.getInstance();
      
      for (GenericWidget<?> w : widgets)
      {
         WidgetConfig cfg = w.config;

         WidgetConfig active = mgr.getActiveConfig(cfg.id);
         if (active == null)
         {
            mgr.addWidgetConfig(cfg);
         }
         else
         {
            assert(cfg.getClass().equals(active.getClass()));
         }
      }
   }

   /**
    * Recursively finds menu-item in this menu tree. It can appear on different levels.
    * Menu-item is searched using some sort of depth-first algorithm.
    * 
    * <pre>
    * 1. Lets assume that cur is the current node in a MENU tree. Initial value is a DEFINE_MENU node.
    * 2. First we are searching the given MENU-ITEM in all immediate child MENU-ITEMs.
    * 3. If we've found the match than return success.
    * 4. If we've not found the match we are going into the all SUB-MENUs and for each of them we are repeating steps from the step 2.
    * 5. If we've searched all tree and didn't find a match - just return null.    
    * </pre>
    * 
    * @param   name
    *          MENU-ITEM name.
    *          
    * @return  concrete {@link MenuItemWidget} or null if not found.
    */
   protected MenuItemWidget findMenuItemInt(String name)
   {
      Collection<GenericWidget<?>> widgets = widgets();
      
      // check MENU-ITEMS of the first level
      for(GenericWidget<?> gw : widgets)
      {
         if (gw instanceof MenuItemWidget && name.equals(gw.name().getValue()))
         {
            return (MenuItemWidget) gw;
         }
      }
      
      // recursively check SUB-MENUs
      for(GenericWidget<?> gw : widgets)
      {
         if (gw instanceof SubMenuWidget)
         {
            MenuItemWidget mi = ((SubMenuWidget)gw).findMenuItemInt(name);
            
            if (mi != null)
            {
               return mi;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Finds sub-menu in this menu tree. It can appear only once. 
    * Simple depth-first search is used.
    * 
    * @param   name
    *          Sub-menu name.
    * 
    * @return  concrete {@link SubMenuWidget} or null if not found.
    */
   protected SubMenuWidget findSubMenuInt(String name)
   {
      for (GenericWidget<?> gw : widgets())
      {
         if (!(gw instanceof SubMenuWidget))
            continue;
         
         if (name.equals(gw.name().getValue()))
            return (SubMenuWidget) gw;
         
         SubMenuWidget sm = ((SubMenuWidget)gw).findSubMenuInt(name);
         
         if (sm != null)
         {
            return sm;
         }
      }
      
      return null;
   }

   /**
    * Delete this menu, if it is not dynamic and its associated instantiating procedure 
    * is no longer exists.
    */
   void menuDelete()
   {
      if (!_dynamic() && instantiatingProcedure().isUnknown())
      {
         // delete all widgets
         for (GenericWidget<?> widget : n2w.values())
         {
            widget.delete();
         }
         
         // force deletion of a static menu for which its instantiating procedure 
         // is no longer available.
         delete();
      }
   }

   /**
    * Create an array of widgets in this frame that are static including possible header ones.
    *
    * @return   The list of widget ids for this frame.
    */
   public int[] getStaticWidgetIds()
   {
      Set<Integer> widgetIds = new LinkedHashSet<Integer>();

      for (GenericWidget<?> gw : widgets)
      {
         if (!gw._dynamic())
         {
            widgetIds.add(gw.getId());
         }
      }
      
      int[] result = new int[widgetIds.size()];
      int i = 0;

      for (Iterator<Integer> it = widgetIds.iterator(); it.hasNext();)
      {
         result[i++] = it.next();
      }

      return result;
   }
   
   /**
    * Accesses the first widget in the list of contained widgets.
    *
    * @return   The first widget or <code>null</code> if no widgets are
    *           contained.
    */
   public handle getFirstChild()
   {
      if (widgets.isEmpty())
      {
         return new handle();
      }
      else
      {
         return new handle(widgets.get(0));
      }
   }

   /**
    * Accesses the last widget in the list of contained widgets.
    *
    * @return   The last widget or <code>null</code> if no widgets are
    *           contained.
    */
   public handle getLastChild()
   {
      if (widgets.isEmpty())
      {
         return new handle();
      }
      else
      {
         return new handle(widgets.get(widgets.size() - 1));
      }
   }
}
