/*
** Module   : MenuDescription.java
** Abstract : contains the configuration of a given menu or family of menus.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------Description-----------------
** 001 VIG 20150124 Creation initial version.
** 002 HC  20150323 WidgetId.getId() renamed to asInt().
** 003 VIG 20150304 Added child descriptions map and correspondent methods.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.io.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.client.*;

/**
 * This is a container for the configuration of a given menu or family of
 * menus which is efficiently serialized to the client from the server.
 */
public class MenuDescription
implements Externalizable
{
   /** The widget configuration data. */
   private LinkedHashMap<Integer, WidgetConfig> widgets = new LinkedHashMap<>();
   
   /** Descriptions of immediate child containers */
   private LinkedHashMap<Integer, MenuDescription> descriptions = new LinkedHashMap<>();

   /** Root menu ID. */
   private int menuId = 0;
   
   /** Suppress redraw on push of menu description. */
   private boolean suppressRedraw = false; 
   
   /**
    * Default constructor (only used for serialization).
    */
   public MenuDescription()
   {
   }
   
   /**
    * The constructor creates a menu description with the given root frame
    * ID.
    * 
    * @param    id
    *           Identifies the root frame (or the only frame if this is not
    *           going to hold the configuration of a frame family).
    */
   public MenuDescription(int id)
   {
      this.menuId = id;
   }
   
   /**
    * Get the root menu ID.
    * 
    * @return   The root menu ID.
    */
   public int getRootMenuID()
   {
      return menuId;
   }
   
   /**
    * Gets an array of all immediate widget IDs in a given menu.
    * 
    * @param    menuId
    *           ID of the menu for which widget IDs are requested.
    * 
    * @return   The list of all widget IDs for the specified menu.
    */
   public ArrayList<Integer> getWidgetsInMenu(int menuId)
   {
      ArrayList<Integer> wids = new ArrayList<Integer>();
      Iterator<WidgetConfig> iter = getConfigsIterator();
 
      while (iter.hasNext())
      {
         WidgetConfig config = iter.next();
         int parentId = 0;
         if (config instanceof SubMenuConfig)
            parentId = ((SubMenuConfig) config).parentId;
         else if (config instanceof MenuItemConfig)
            parentId = ((MenuItemConfig) config).parentId;
         
         if (parentId == menuId)
         {
            wids.add(new Integer(config.id.asInt()));
         }
      }

      return wids;
   }
   
   /**
    * Get an iterator to enumerate all widget configurations inside this
    * menu description.
    * 
    * @return   The iterator.
    */
   public Iterator<WidgetConfig> getConfigsIterator()
   {
      return widgets.values().iterator();
   }

   /**
    * Get an array of all sub-descriptions inside this menu description.
    * 
    * @return   The array.
    */
   public MenuDescription[] getChildDescrArray()
   {
      return descriptions.values().toArray(new MenuDescription[descriptions.size()]);
   }
   
   /**
    * Gets a specific widget configuration instance for the widget with the
    * ID provided.
    * 
    * @param    id
    *           The widget ID.
    * 
    * @return   The requested widget configuration object if exists,
    *           <code>null</code> otherwise.
    */
   public WidgetConfig getConfig(int id)
   {
      return widgets.get(new Integer(id));
   }
   
   /**
    * Add a widget configuration to the current menu description.
    * 
    * @param    config
    *           New config to be added to the current menu description.
    */
   public void addConfig(WidgetConfig config)
   {
      addConfigInt(config);
   }

   /**
    * Add a widget configuration (to the current menu description) and
    * assign that widget the specified ID.
    * 
    * @param    config
    *           New config to be added to the current menu description.
    * @param    widgetID
    *           Widget ID.
    */
   public void addConfig(WidgetConfig config, int widgetID)
   {
      config.id = new WidgetId(widgetID);
      addConfigInt(config);
   }
   
   /**
    * Add a widget configuration (to the current menu description) and
    * assign that widget the specified ID.
    * 
    * @param    description
    *           New child container description
    * @param    widgetID
    *           Child container ID.
    */
   public void addChildDescription(MenuDescription description, int widgetID)
   {
      if (descriptions.containsKey(widgetID))
      {
         throw new RuntimeException("Duplicate description for ID = " + widgetID);
      }
      
      if (!widgets.containsKey(widgetID))
      {
         throw new RuntimeException("No correspondent widget for ID = " + widgetID);
      }
      
      descriptions.put(widgetID, description);
   }
   
   /**
    * Removes a widget configuration from the current menu description.
    * 
    * @param    config
    *           The config to be removed from the current menu description.
    */
   public void removeConfig(WidgetConfig config)
   {
      int wid = config.id == null ? -1 : config.id.asInt();
      
      if (widgets.get(wid) == null)
      {
         throw new RuntimeException("Missing widget for remove, ID = " + wid);
      }
      
      widgets.remove(wid);
      
      if (descriptions.containsKey(wid))
      {
         descriptions.remove(wid);
      }
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If pay-load can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      menuId         = in.readInt();    
      suppressRedraw = in.readBoolean();
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), widgets);
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), descriptions);
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(menuId);
      out.writeBoolean(suppressRedraw);
      out.writeObject(WidgetEntry.mapToArray(widgets));
      out.writeObject(WidgetEntry.mapToArray(descriptions));
   }   
   
   /**
    * Set new value for 'suppress redraw' flag.
    * 
    * @param   suppressRedraw
    *          New value of flag.
    */
   public void setSuppressRedraw(boolean suppressRedraw)
   {
      this.suppressRedraw = suppressRedraw; 
   }

   /**
    * Get current value of 'suppress redraw' flag.
    * 
    * @return  The current value of the flag. 
    */
   public boolean isSuppressRedraw()
   {
      return suppressRedraw; 
   }

   /**
    * Common worker to add a widget configuration to the menu description.
    * 
    * @param    config
    *           Widget configuration.
    */
   private void addConfigInt(WidgetConfig config)
   {
      int wid = (config.id == null ? -1 : config.id.asInt());
      
      if (wid == -1)
      {
         assert (!SessionManager.get().isLeaf());
         
         // allocate an ID
         // we must never be here on the client side
         WidgetId nextID = WidgetId.nextID();
         config.id = nextID;
         wid = nextID.asInt();
      }
      
      Integer iwid = new Integer(wid);
      
      if (widgets.get(iwid) != null)
      {
         throw new RuntimeException("Duplicate widget ID = " + iwid);
      }
      
      widgets.put(iwid, config);  
   }
}
