/*
** Module   : MessageAreaImpl.java
** Abstract : MESSAGE-AREA widget, for ChUI clients.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140711 Created initial version
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 MAG 20141007 Changed to extends AbstractMessageArea.
** 007 MAG 20141009 Code from AbstractMessageArea was moved here because is ChUI specific only.
**                  AbstractMessageArea class has benn deleted.
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 HC  20150323 Removed business logic from config classes.
** 010 VIG 20150407 Fixed typified Window.getDefaultWindow method call.
*/

package com.goldencode.p2j.ui.client.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Widget associated with the window's message area.
 * <p>
 * For ChUI clients, it always holds 2 message lines.
 */
@SuppressWarnings("unchecked")
public class MessageAreaImpl
extends AbstractWidget<ChuiOutputManager> 
implements MessageArea<ChuiOutputManager>,
           WidgetWithConfig<BaseConfig>
{
   /** Two lines which render message text. */ 
   protected List<StatusLine<ChuiOutputManager>> messageLines = null;
   
   /** The number of lines in this message area. */
   private int lines;

   /** Widget configuration */
   protected BaseConfig config;
   
   /**
    * Initialize this message-area.
    * 
    * @param   id
    *          The widget's ID.
    * @param   lines
    *          Number of lines in message area.
    */
   public MessageAreaImpl(WidgetId id, int lines)
   {
      this.lines = lines;
      this.messageLines = new ArrayList<>();
      
      for (int i = 0; i < lines; i++)
      {
         this.messageLines.add(new MessageLine(WidgetId.nextID()));
      }
      
      this.config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @Override
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Attach this widget to the specified window.
    * 
    * @param    window
    *           The parent window.
    */
   @Override
   public void attach(Window<ChuiOutputManager> window)
   {
      Dimension screenSize = window.screenDimension();
      for (int i = 0; i < 2; i++)
      {
         StatusLine<ChuiOutputManager> line = messageLines.get(i);
         window.getContentPane().add(line);
         line.setLocation(0, screenSize.height - 3 + i);
         line.setLength(screenSize.width);
      }
      
      this.setParent(window);
   }

   /**
    * Get all the message texts in this message-area.
    * 
    * @return   See above.
    */
   @Override
   public String[] getMessageTexts()
   {
      String[] msgs = new String[lines];
      for (int i = 0; i < lines; i++)
      {
         msgs[i] = messageLines.get(i).getText();
      }
      
      return msgs;
   }
   
   /**
    * Get all the message colors in this message-area.
    * 
    * @return   See above.
    */
   @Override
   public Color[] getMessageColors()
   {
      Color[] colors = new Color[lines];
      for (int i = 0; i < lines; i++)
      {
         colors[i] = messageLines.get(i).getColor();
      }
      
      return colors;
   }

   /**
    * Force widget visibility change without associated repainting.
    * 
    * @param   visibility
    *          New visibility state.
    */
   @Override
   public void forceVisibility(boolean visibility)
   {
      super.forceVisibility(visibility);
      
      for (StatusLine<ChuiOutputManager> msg : messageLines)
      {
         msg.forceVisibility(visibility);
      }
   }
   
   /**
    * Shows the new message text.
    * 
    * @param   message
    *          Array of two text message which need to be displayed inside the
    *          message area.
    * @param   cs
    *          Array of two color specifications.
    */
   @Override
   public void setMessageText(String[] message, Color[] cs)
   {
      setMessageText(message, cs, false);
   }
   
   /**
    * Set message main worker routine.
    * 
    * @param   msg
    *          Message text array.
    * @param   cs
    *          Message color array.
    * @param   force
    *          <code>true</code> means force redisplaying of messages even
    *          if message text is already <code>null</code>.
    */
   @Override
   public void setMessageText(String[] msg, Color[] cs, boolean force)
   {
      if (msg == null || msg.length < 2)
      {
         msg = new String[] {null, null};
      }
      
      StatusLine<ChuiOutputManager> line0 = messageLines.get(0);
      StatusLine<ChuiOutputManager> line1 = messageLines.get(1);
      
      if (line0.getText() == null && 
          line1.getText() == null &&
          msg[0] == null                   &&
          msg[1] == null                   && 
          !force)
      {
         // do not force cleanup
         return;
      }
   
      // save the bitmap
      ScreenBitmap bitmap = getAndResetBitmap();
   
      EventManager.pushEventQueue();
      
      try
      {
         Window<ChuiOutputManager> window = Window.<ChuiOutputManager>getDefaultWindow();

         ArrayList<Frame<?>> list = new ArrayList<>();
         boolean drawSecondLine = (line1.getText() != null || msg[1] != null);
         
         Rectangle rect = window.getMessageLineRectangle(0);
         
         if (drawSecondLine)
            rect = rect.union(window.getMessageLineRectangle(1));
         
         Window.findFrames(list, rect, true, false);
         
         Container<ChuiOutputManager> c = Window.<ChuiOutputManager>getDefaultWindow().getContentPane();
         
         c.moveToTop(line0);
         c.moveToTop(line1);
   
         line0.setText(msg[0]);
         line1.setText(msg[1]);
         
         if (cs != null && cs.length == 2)
         {
            for (int i = 0; i < cs.length; i++)
            {
               if(cs[i] != null)
                  messageLines.get(i).setColor(cs[i]);
            }
         }
   
         // reset the screen
         // add the new clipping rectangle
         resetBitmap();
         clipWith(rect);
         
         line0.draw();
         
         if (drawSecondLine)
            line1.draw();
      }
      
      // cleanup and restore state
      finally
      {
         // ignore all events if they were generated
         while (EventManager.getEvent() != null)
         {
         }
         EventManager.popEventQueue();
         
         // restore the bitmap
         setBitmap(bitmap);
         sync();
      }
   }
   
   /**
    * This is a no-op, and it must not be called for ChUI clients.  In ChUI clients, current
    * implementation treats each message line individually.
    */
   @Override
   public void draw()
   {
      throw new RuntimeException("This API must not be used, for this widget implementation!");
   }

   /**
    * This is a no-op, and it must not be called for ChUI clients.  In ChUI clients, current
    * implementation treats each message line individually.
    */
   @Override
   public Dimension dimension()
   {
      throw new RuntimeException("This API must not be used, for this widget implementation!");
   }

   /**
    * This is a no-op, and it must not be called for ChUI clients.  In ChUI clients, current
    * implementation treats each message line individually.
    */
   @Override
   public double width()
   {
      throw new RuntimeException("This API must not be used, for this widget implementation!");
   }

   /**
    * This is a no-op, and it must not be called for ChUI clients.  In ChUI clients, current
    * implementation treats each message line individually.
    */
   @Override
   public double height()
   {
      throw new RuntimeException("This API must not be used, for this widget implementation!");
   }

   /**
    * This is a no-op, and it must not be called for ChUI clients.  In ChUI clients, current
    * implementation treats each message line individually.
    */
   @Override
   public Dimension minimumSize()
   {
      throw new RuntimeException("This API must not be used, for this widget implementation!");
   }
}
